@@ -10,6 +10,7 @@
 #include <unistd.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 static int Sflag;
 
@@ -32,7 +33,7 @@ export(char *file)
 	}
 
 	char from[1024] = "nobody";
-	time_t date = -1;
+	time_t date = 0;
 
 	if (fseek(infile, 0L, SEEK_SET) && errno == ESPIPE) {
 		date = time(0);
@@ -67,7 +68,7 @@ export(char *file)
 	char *line = 0;
 	size_t linelen = 0;
 
-	printf("From %s %s", from, ctime(&date));
+	printf("From %s %s", from, asctime(gmtime(&date)));
 
 	int in_header = 1;
 	int final_nl = 0;
@@ -124,6 +125,9 @@ export(char *file)
 	if (!final_nl)
 		putchar('\n');
 
+	// ensure empty line at end of message
+	putchar('\n');
+
 	fclose(infile);
 }
 
@@ -141,6 +145,8 @@ main(int argc, char *argv[])
 
 	status = 0;
 
+	xpledge("stdio rpath", "");
+
 	if (argc == optind && isatty(0))
 		blaze822_loop1(":", export);
 	else
