@@ -7,9 +7,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <unistd.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 static char *hflag;
 static char *pflag;
@@ -58,26 +60,51 @@ headerall(struct message *msg)
 	blaze822_free(msg);
 }
 
+void
+print_quoted(char *s)
+{
+	char *t;
+
+	for (t = s; *t; t++)
+                if ((unsigned char)*t < 32 || strchr("()<>[]:;@\\,.\"", *t))
+			goto quote;
+
+	printf("%s", s);
+	return;
+
+quote:
+	putchar('"');
+	for (t = s; *t; t++) {
+		if (*t == '"' || *t == '\\')
+			putchar('\\');
+		putchar(*t);
+	}
+	putchar('"');
+
+}
+
 void
 print_addresses(char *s)
 {
 	char *disp, *addr;
-	char sdec[4096];
-
-	if (dflag) {
-		blaze822_decode_rfc2047(sdec, s, sizeof sdec, "UTF-8");
-		sdec[sizeof sdec - 1] = 0;
-		s = sdec;
-	}
+	char ddec[4096];
 
 	while ((s = blaze822_addr(s, &disp, &addr))) {
 		if (Hflag && addr)
 			printf("%s\t", curfile);
 
-		if (disp && addr)
-			printf("%s <%s>\n", disp, addr);
-		else if (addr)
+		if (disp && addr) {
+			if (dflag) {
+				blaze822_decode_rfc2047(ddec, disp, sizeof ddec, "UTF-8");
+				ddec[sizeof ddec - 1] = 0;
+				disp = ddec;
+			}
+
+			print_quoted(disp);
+			printf(" <%s>\n", addr);
+		} else if (addr) {
 			printf("%s\n", addr);
+		}
 	}
 }
 
@@ -138,7 +165,7 @@ headermany(struct message *msg)
 				*n = 0;
 
 			size_t l = strlen(h);
-			if (strncmp(hdr, h, l) == 0 && hdr[l] == ':') {
+			if (strncasecmp(hdr, h, l) == 0 && hdr[l] == ':') {
 				hdr += l + 1;
 				while (*hdr == ' ' || *hdr == '\t')
 					hdr++;
@@ -220,6 +247,8 @@ main(int argc, char *argv[])
 
 	status = 1;
 
+	xpledge("stdio rpath", "");
+
 	if (argc == optind && isatty(0))
 		blaze822_loop1(".", header);
 	else
