@@ -1,6 +1,7 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
+#include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -15,6 +16,7 @@
 #include <unistd.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 static int cflag;
 static int rflag;
@@ -57,6 +59,11 @@ int gen_b64(uint8_t *s, off_t size)
 	return 0;
 }
 
+#define qphrasevalid(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'Z') || \
+                         (c >= 'a' && c <= 'z') || \
+                         c == '!' || c == '*' || c == '+' || c == '-' || \
+                         c == '/')
+
 size_t
 gen_qp(uint8_t *s, off_t size, size_t maxlinelen, size_t linelen)
 {
@@ -121,6 +128,10 @@ gen_qp(uint8_t *s, off_t size, size_t maxlinelen, size_t linelen)
 			putc_unlocked('\n', stdout);
 			linelen = 0;
 			prev = 0;
+		} else if (header && !qphrasevalid(s[i])) {
+			printf("=%02X", s[i]);
+			linelen += 3;
+			prev = '_';
 		} else {
 			putc_unlocked(s[i], stdout);
 			linelen++;
@@ -246,7 +257,7 @@ gen_file(char *file, char *ct)
 	}
 
 	if (bitlow == 0 && bithigh == 0 &&
-	    maxlinelen <= 78 && content[size-1] == '\n') {
+	    maxlinelen <= 78) {
 		if (!ct)
 			ct = "text/plain";
 		printf("Content-Type: %s\n", ct);
@@ -282,6 +293,9 @@ print_header(char *line) {
 	char *s, *e;
 	size_t l = strlen(line);
 
+	if (l == 0)
+		return;
+
 	if (line[l-1] == '\n')
 		line[l-1] = 0;
 
@@ -298,17 +312,35 @@ print_header(char *line) {
 	}
 
 	int prevq = 0;  // was the previous word encoded as qp?
+	char prevqs = 0;  // was the previous word a quoted-string?
 
-	size_t linelen = s - line;
+	ssize_t linelen = s - line;
 
 	while (*s) {
 		size_t highbit = 0;
+		int qs = 0;
 		e = s;
 		while (*e && *e == ' ')
 			e++;
-		for (; *e && *e != ' '; e++) {
-			if ((uint8_t)*e >= 127)
-				highbit++;
+
+		if (*e == '"') {  // scan quoted-string, encode at once
+			s = e;
+			for (e++; *e && *e != '"'; e++) {
+				if (*e == '\\')
+					e++;
+				if ((uint8_t)*e >= 127)
+					highbit++;
+			}
+			if (*e == '"')
+				e++;
+			qs = 1;
+		} else {  // scan word
+			while (*e && *e == ' ')
+				e++;
+			for (; *e && *e != ' '; e++) {
+				if ((uint8_t)*e >= 127)
+					highbit++;
+			}
 		}
 
 		if (!highbit) {
@@ -323,7 +355,7 @@ print_header(char *line) {
 				// space at beginning of line
 				goto force_qp;
 			}
-			if (*s != ' ') {
+			if (*s != ' ' && !(prevqs && !prevq && *(s-1) != ' ')) {
 				printf(" ");
 				linelen++;
 			}
@@ -334,6 +366,12 @@ print_header(char *line) {
 force_qp:
 			if (!prevq && *s == ' ')
 				s++;
+
+			if (qs && *s == '"')
+				s++;
+			if (qs && e > s && *(e-1) == '"')
+				e--;
+
 			if (linelen >= 78 - 13 - 4 ||
 			    (e-s < (78 - 13)/3 &&
 			     e-s >= (78 - linelen - 13)/3)) {
@@ -347,12 +385,33 @@ force_qp:
 			printf("?=");
 			linelen += 2;
 			prevq = 1;
+
+			if (qs && *e == '"')
+				e++;
 		}
+		prevqs = qs;
 		s = e;
 	}
 	printf("\n");
 }
 
+static int
+valid_content_type(char *s)
+{
+	int slash = 0;
+
+	for (; *s; s++)
+		if (*s == '/')
+			slash++;
+		else if (isalnum(*s) || *s == '-' || *s == '+' || *s == '.' ||
+		    *s == ';' || *s == '=' || *s == '#')
+			; /* ok */
+		else
+			return 0;
+
+	return slash == 1;
+}
+
 int
 gen_build()
 {
@@ -364,30 +423,45 @@ gen_build()
 	size_t linelen = 0;
 	int inheader = 1;
 	int intext = 0;
+	int emptybody = 1;
+	int ret = 0;
+	char *contenttype = 0;
+	char *contenttransferenc = 0;
 
 	while (1) {
 		ssize_t read = getdelim(&line, &linelen, '\n', stdin);
 		if (read == -1) {
-			if (feof(stdin))
-				break;
-			else
+			if (feof(stdin)) {
+				if (!emptybody)
+					break;
+				line = strdup(inheader ? "\n" : "");
+			} else { // errored
 				exit(1);
+			}
 		}
 		if (inheader) {
 			if (line[0] == '\n') {
 				inheader = 0;
 				printf("MIME-Version: 1.0\n");
 				if (rflag) {
-					printf("Content-Type: text/plain; charset=UTF-8\n");
-					printf("Content-Transfer-Encoding: quoted-printable\n\n");
-
+					printf("Content-Type:%s", contenttype ? contenttype : " text/plain; charset=UTF-8\n");
+					printf("Content-Transfer-Encoding:%s", contenttransferenc ? contenttransferenc : " quoted-printable\n");
+					printf("\n");
 				} else {
 					printf("Content-Type: %s; boundary=\"%s\"\n", tflag, sep);
 					printf("\n");
 					printf("This is a multipart message in MIME format.\n");
 				}
 			} else {
-				print_header(line);
+				if (strncasecmp(line, "Content-Type:", 13) == 0) {
+					free(contenttype);
+					contenttype = strdup(line+13);
+				} else if (strncasecmp(line, "Content-Transfer-Encoding:", 26) == 0) {
+					free(contenttransferenc);
+					contenttransferenc = strdup(line+26);
+				} else {
+					print_header(line);
+				}
 			}
 			continue;
 		}
@@ -397,12 +471,14 @@ gen_build()
 			if (f) {
 				char of = *f;
 				*f = 0;
-				if (strchr(line, '/')) {
+				if (valid_content_type(line+1)) {
 					printf("\n--%s\n", sep);
 					if (line[read-1] == '\n')
 						line[read-1] = 0;
-					gen_file(f+1, (char *)line+1);
+					if (gen_file(f+1, line+1) != 0)
+						ret = 1;
 					intext = 0;
+					emptybody = 0;
 					continue;
 				}
 				*f = of;
@@ -411,20 +487,26 @@ gen_build()
 
 		if (!rflag && !intext) {
 			printf("\n--%s\n", sep);
-			printf("Content-Type: text/plain; charset=UTF-8\n");
+			printf("Content-Type:%s", contenttype ? contenttype : " text/plain; charset=UTF-8\n");
 			printf("Content-Disposition: inline\n");
-			printf("Content-Transfer-Encoding: quoted-printable\n\n");
+			printf("Content-Transfer-Encoding:%s", contenttransferenc ? contenttransferenc : " quoted-printable\n");
+			printf("\n");
 
 			intext = 1;
 		}
 
-		gen_qp((uint8_t *)line, strlen(line), 78, 0);
+		if (contenttransferenc)
+			printf("%s", line);
+		else
+			gen_qp((uint8_t *)line, strlen(line), 78, 0);
+
+		emptybody = 0;
 	}
 	if (!rflag && !inheader)
 		printf("\n--%s--\n", sep);
 
 	free(line);
-	return 0;
+	return ret;
 }
 
 int
@@ -500,6 +582,8 @@ usage:
 	if (argc != optind)
 		goto usage;
 
+	xpledge("stdio rpath", "");
+
 	if (cflag)
 		return check();
 
