@@ -23,6 +23,7 @@
 #endif
 
 #include <sys/ioctl.h>
+#include <sys/mman.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 
@@ -42,10 +43,12 @@
 #include <wchar.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 enum op {
 	EXPR_OR = 1,
 	EXPR_AND,
+	EXPR_COND,
 	EXPR_NOT,
 	EXPR_LT,
 	EXPR_LE,
@@ -61,9 +64,12 @@ enum op {
 	EXPR_REGEXI,
 	EXPR_PRUNE,
 	EXPR_PRINT,
+	EXPR_REDIR_FILE,
+	EXPR_REDIR_PIPE,
 	EXPR_TYPE,
 	EXPR_ALLSET,
 	EXPR_ANYSET,
+	EXPR_BINDING,
 };
 
 enum prop {
@@ -76,11 +82,12 @@ enum prop {
 	PROP_REPLIES,
 	PROP_SIZE,
 	PROP_TOTAL,
-	PROP_FROM,
-	PROP_TO,
 	PROP_INDEX,
 	PROP_DATE,
 	PROP_FLAG,
+	PROP_HDR,
+	PROP_HDR_ADDR,
+	PROP_HDR_DISP,
 };
 
 enum flags {
@@ -113,11 +120,13 @@ struct expr {
 		int64_t num;
 		regex_t *regex;
 		enum var var;
-	} a, b;
+		struct binding *binding;
+	} a, b, c;
 	int extra;
 };
 
 struct mailinfo {
+	char *file;
 	char *fpath;
 	struct stat *sb;
 	struct message *msg;
@@ -143,6 +152,35 @@ struct thread {
 	struct mlist *cur;
 };
 
+struct file {
+	enum op op;
+	const char *name;
+	const char *mode;
+	FILE *fp;
+	struct file *next;
+};
+
+struct pos {
+	char *pos;
+	char *line;
+	size_t linenr;
+};
+
+struct binding {
+	char *name;
+	size_t nlen;
+	size_t refcount;
+	struct expr *expr;
+	struct binding *next;
+};
+
+struct scope {
+	struct binding *bindings;
+	struct scope *prev;
+};
+
+struct scope *scopeq = NULL;
+
 static struct thread *thr;
 
 static char *argv0;
@@ -156,15 +194,54 @@ static long num;
 static struct expr *expr;
 static long cur_idx;
 static char *cur;
-static char *pos;
 static time_t now;
 static int prune;
 
+static char *pos;
+static const char *fname;
+static char *line = NULL;
+static int linenr = 0;
+
+static struct file *files, *fileq = NULL;
+
+static void *
+xcalloc(size_t nmemb, size_t size)
+{
+	void *r;
+	if ((r = calloc(nmemb, size)))
+		return r;
+	perror("calloc");
+	exit(2);
+}
+
+static char *
+xstrdup(const char *s)
+{
+	char *r;
+	if ((r = strdup(s)))
+		return r;
+	perror("strdup");
+	exit(2);
+}
+
 static void
 ws()
 {
-	while (isspace((unsigned char)*pos))
-		pos++;
+	for (; *pos;) {
+		while (isspace((unsigned char)*pos)) {
+			if (*pos == '\n') {
+				line = pos+1;
+				linenr++;
+			}
+			pos++;
+		}
+		if (*pos != '#')
+			break;
+
+		pos += strcspn(pos, "\n\0");
+		if (*pos != '\n')
+			break;
+	}
 }
 
 static int
@@ -179,17 +256,45 @@ token(char *token)
 	}
 }
 
+static int
+peek(char *token)
+{
+	return strncmp(pos, token, strlen(token)) == 0;
+}
+
 noreturn static void
 parse_error(char *msg, ...)
 {
 	va_list ap;
 	va_start(ap, msg);
-	fprintf(stderr, "%s: parse error: ", argv0);
+	fprintf(stderr, "%s: parse error: %s:%d:%ld: ", argv0, fname, linenr, pos-line+1);
 	vfprintf(stderr, msg, ap);
 	fprintf(stderr, "\n");
 	exit(2);
 }
 
+noreturn static void
+parse_error_at(struct pos *savepos, char *msg, ...)
+{
+	char *e;
+
+	if (savepos) {
+		pos = savepos->pos;
+		line = savepos->line;
+		linenr = savepos->linenr;
+	}
+
+	va_list ap;
+	va_start(ap, msg);
+	fprintf(stderr, "%s: parse error: %s:%d:%ld: ", argv0, fname, linenr, pos-line+1);
+	vfprintf(stderr, msg, ap);
+	fprintf(stderr, " at '");
+	for (e = pos+15; *pos && *pos != '\n' && pos <= e; pos++)
+		putc(*pos, stderr);
+	fprintf(stderr, "'\n");
+	exit(2);
+}
+
 static struct expr *
 mkexpr(enum op op)
 {
@@ -201,6 +306,58 @@ mkexpr(enum op op)
 	return e;
 }
 
+static void
+freeexpr(struct expr *e)
+{
+	if (!e) return;
+	switch (e->op) {
+	case EXPR_OR:
+	case EXPR_AND:
+		freeexpr(e->a.expr);
+		freeexpr(e->b.expr);
+		break;
+	case EXPR_COND:
+		freeexpr(e->a.expr);
+		freeexpr(e->b.expr);
+		freeexpr(e->c.expr);
+		break;
+	case EXPR_NOT:
+		freeexpr(e->a.expr);
+		break;
+	case EXPR_BINDING:
+		e->a.binding->refcount -= 1;
+		if (e->a.binding->refcount == 0) {
+			freeexpr(e->a.binding->expr);
+			free(e->a.binding);
+		}
+		break;
+	case EXPR_REDIR_FILE:
+	case EXPR_REDIR_PIPE:
+		free(e->a.string);
+		free(e->b.string);
+		break;
+	case EXPR_STREQ:
+	case EXPR_STREQI:
+	case EXPR_GLOB:
+	case EXPR_GLOBI:
+	case EXPR_REGEX:
+	case EXPR_REGEXI:
+		switch (e->a.prop) {
+		case PROP_PATH: break;
+		case PROP_HDR:
+		case PROP_HDR_ADDR:
+		case PROP_HDR_DISP:
+			free(e->b.string);
+		}
+		if (e->op == EXPR_REGEX || e->op == EXPR_REGEXI)
+			regfree(e->c.regex);
+		else
+			free(e->c.string);
+		break;
+	}
+	free(e);
+}
+
 static struct expr *
 chain(struct expr *e1, enum op op, struct expr *e2)
 {
@@ -244,33 +401,136 @@ parse_op()
 	return 0;
 }
 
+static int
+parse_ident(char **sp, size_t *lp)
+{
+	char *p = pos;
+	if (!isalpha(*pos) && *pos != '_')
+		return 0;
+	p++;
+	while (*p && (isalnum(*p) || *p == '_'))
+		p++;
+	*sp = pos;
+	*lp = p-pos;
+	pos = p;
+	ws();
+	return 1;
+}
+
+static struct expr *
+parse_binding()
+{
+	struct scope *sc;
+	struct binding *b;
+	char *s;
+	size_t l = 0;
+	struct pos savepos = { pos, line, linenr };
+
+	if (parse_ident(&s, &l)) {
+		for (sc = scopeq; sc; sc = sc->prev) {
+			for (b = sc->bindings; b; b = b->next) {
+				if (b->nlen == l && strncmp(b->name, s, l) == 0) {
+					struct expr *e = mkexpr(EXPR_BINDING);
+					e->a.binding = b;
+					b->refcount++;
+					return e;
+				}
+			}
+		}
+	}
+
+	parse_error_at(&savepos, "unknown expression");
+	return 0;
+}
+
+static struct expr *parse_cond();
+
+static struct expr *
+parse_let()
+{
+	if (!token("let"))
+		return parse_binding();
+
+	struct scope *sc;
+	char *s;
+	size_t l;
+
+	sc = xcalloc(1, sizeof (struct scope));
+	sc->prev = scopeq;
+	scopeq = sc;
+
+	struct binding *b, *bq;
+	struct expr *e;
+	bq = NULL;
+	for (;;) {
+		if (!parse_ident(&s, &l))
+			parse_error_at(NULL, "missing ident");
+		if (!token("="))
+			parse_error_at(NULL, "missing =");
+		e = parse_cond();
+		b = xcalloc(1, sizeof (struct binding));
+		b->name = s;
+		b->nlen = l;
+		b->expr = e;
+		if (bq) {
+			bq->next = b;
+			bq = b;
+		} else {
+			bq = sc->bindings = b;
+		}
+		if (!token("let"))
+			break;
+	}
+	if (!token("in"))
+		parse_error_at(NULL, "missing `in`");
+
+	e = parse_cond();
+
+	struct binding *bs;
+	for (b = sc->bindings; b; b = bs) {
+		bs = b->next;
+		if (b->refcount != 0)
+			continue;
+		freeexpr(b->expr);
+		free(b);
+	}
+
+	scopeq = sc->prev;
+	free(sc);
+
+	return e;
+}
+
 static struct expr *parse_cmp();
-static struct expr *parse_or();
 
 static struct expr *
 parse_inner()
 {
 	if (token("prune")) {
-		struct expr *e = mkexpr(EXPR_PRUNE);
-		return e;
+		return mkexpr(EXPR_PRUNE);
 	} else if (token("print")) {
+		return mkexpr(EXPR_PRINT);
+	} else if (token("skip")) {
 		struct expr *e = mkexpr(EXPR_PRINT);
-		return e;
+		struct expr *not = mkexpr(EXPR_NOT);
+		not->a.expr = e;
+		return not;
 	} else if (token("!")) {
 		struct expr *e = parse_cmp();
 		struct expr *not = mkexpr(EXPR_NOT);
 		not->a.expr = e;
 		return not;
-	} else if (token("(")) {
-		struct expr *e = parse_or();
+	}
+	if (peek("(")) {
+		struct pos savepos = { pos, line, linenr };
+		(void) token("(");
+		struct expr *e = parse_cond();
 		if (token(")"))
 			return e;
-		parse_error("missing ) at '%.15s'", pos);
-		return 0;
-	} else {
-		parse_error("unknown expression at '%.15s'", pos);
+		parse_error_at(&savepos, "unterminated (");
 		return 0;
 	}
+	return parse_let();
 }
 
 static int
@@ -284,7 +544,7 @@ parse_string(char **s)
 		pos++;
 		while (*pos &&
 		    (*pos != '"' || (*pos == '"' && *(pos+1) == '"'))) {
-			if (len >= bufsiz) {
+			if (len+1 >= bufsiz) {
 				bufsiz = 2*bufsiz + 16;
 				buf = realloc(buf, bufsiz);
 				if (!buf)
@@ -300,7 +560,7 @@ parse_string(char **s)
 			buf[len] = 0;
 		pos++;
 		ws();
-		*s = buf ? buf : "";
+		*s = buf ? buf : xstrdup("");
 		return 1;
 	} else if (*pos == '$') {
 		char t;
@@ -316,6 +576,7 @@ parse_string(char **s)
 		*s = getenv(e);
 		if (!*s)
 			*s = "";
+		*s = xstrdup(*s);
 		*pos = t;
 		ws();
 		return 1;
@@ -337,14 +598,25 @@ parse_strcmp()
 	negate = 0;
 
 	if (token("from"))
-		prop = PROP_FROM;
+		h = xstrdup("from");
 	else if (token("to"))
-		prop = PROP_TO;
+		h = xstrdup("to");
 	else if (token("subject"))
-		h = "subject";
+		h = xstrdup("subject");
+	else if (token("path"))
+		prop = PROP_PATH;
 	else if (!parse_string(&h))
 		return parse_inner();
 
+	if (!prop) {
+		if (token(".")) {
+			if (token("addr"))      prop = PROP_HDR_ADDR;
+			else if (token("disp")) prop = PROP_HDR_DISP;
+			else parse_error_at(NULL, "unknown decode parameter");
+		} else
+			prop = PROP_HDR;
+	}
+
 	if (token("~~~"))
 		op = EXPR_GLOBI;
 	else if (token("~~"))
@@ -372,46 +644,39 @@ parse_strcmp()
 	else if (token("!==") || token("!="))
 		negate = 1, op = EXPR_STREQ;
 	else
-		parse_error("invalid string operator at '%.15s'", pos);
+		parse_error_at(NULL, "invalid string operator");
 
 	char *s;
 	if (!parse_string(&s)) {
-		parse_error("invalid string at '%.15s'", pos);
+		parse_error_at(NULL, "invalid string");
 		return 0;
 	}
 
 	int r = 0;
 	struct expr *e = mkexpr(op);
-
-	if (prop)
-		e->a.prop = prop;
-	else
-		e->a.string = h;
-
-	if (prop == PROP_FROM || prop == PROP_TO) {
-		char *disp, *addr;
-		blaze822_addr(s, &disp, &addr);
-		if (!disp && !addr)
-			parse_error("invalid address at '%.15s'", pos);
-		s = strdup((disp) ? disp : addr);
-		e->extra = (disp) ? 0 : 1;
+	e->a.prop = prop;
+	switch (prop) {
+	case PROP_HDR:
+	case PROP_HDR_ADDR:
+	case PROP_HDR_DISP:
+		e->b.string = h;
+		break;
+	case PROP_PATH: break;
 	}
 
-	if (op == EXPR_REGEX) {
-		e->b.regex = malloc(sizeof (regex_t));
-		r = regcomp(e->b.regex, s, REG_EXTENDED | REG_NOSUB);
-	} else if (op == EXPR_REGEXI) {
-		e->b.regex = malloc(sizeof (regex_t));
-		r = regcomp(e->b.regex, s, REG_EXTENDED | REG_NOSUB | REG_ICASE);
+	if (op == EXPR_REGEX || op == EXPR_REGEXI) {
+		e->c.regex = malloc(sizeof (regex_t));
+		r = regcomp(e->c.regex, s, REG_EXTENDED | REG_NOSUB |
+		    (op == EXPR_REGEXI ? REG_ICASE : 0));
+		if (r != 0) {
+			char msg[256];
+			regerror(r, e->c.regex, msg, sizeof msg);
+			parse_error("invalid regex '%s': %s", s, msg);
+			exit(2);
+		}
+		free(s);
 	} else {
-		e->b.string = s;
-	}
-
-	if (r != 0) {
-		char msg[256];
-		regerror(r, e->b.regex, msg, sizeof msg);
-		parse_error("invalid regex '%s': %s", s, msg);
-		exit(2);
+		e->c.string = s;
 	}
 
 	if (negate) {
@@ -511,7 +776,7 @@ parse_cmp()
 		return parse_flag();
 
 	if (!(op = parse_op()))
-		parse_error("invalid comparison at '%.15s'", pos);
+		parse_error_at(NULL, "invalid comparison");
 
 	int64_t n;
 	if (parse_num(&n)) {
@@ -543,20 +808,20 @@ parse_dur(int64_t *n)
 			parse_error("can't stat file '%s': %s",
 			    s, strerror(errno));
 		*n = st.st_mtime;
-		return 1;
+		goto ret;
 	}
 
 	struct tm tm = { 0 };
 	r = strptime(s, "%Y-%m-%d %H:%M:%S", &tm);
 	if (r && !*r) {
 		*n = mktime(&tm);
-		return 1;
+		goto ret;
 	}
 	r = strptime(s, "%Y-%m-%d", &tm);
 	if (r && !*r) {
 		tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
 		*n = mktime(&tm);
-		return 1;
+		goto ret;
 	}
 	r = strptime(s, "%H:%M:%S", &tm);
 	if (r && !*r) {
@@ -565,7 +830,7 @@ parse_dur(int64_t *n)
 		tm.tm_mon = tmnow->tm_mon;
 		tm.tm_mday = tmnow->tm_mday;
 		*n = mktime(&tm);
-		return 1;
+		goto ret;
 	}
 	r = strptime(s, "%H:%M", &tm);
 	if (r && !*r) {
@@ -575,39 +840,40 @@ parse_dur(int64_t *n)
 		tm.tm_mday = tmnow->tm_mday;
 		tm.tm_sec = 0;
 		*n = mktime(&tm);
-		return 1;
+		goto ret;
 	}
 
 	if (*s == '-') {
-		s++;
-
-		errno = 0;
 		int64_t d;
-		d = strtol(s, &r, 10);
+		errno = 0;
+		d = strtol(s+1, &r, 10);
 		if (errno == 0 && r[0] == 'd' && !r[1]) {
 			struct tm *tmnow = localtime(&now);
 			tmnow->tm_mday -= d;
 			tmnow->tm_hour = tmnow->tm_min = tmnow->tm_sec = 0;
 			*n = mktime(tmnow);
-			return 1;
+			goto ret;
 		}
 		if (errno == 0 && r[0] == 'h' && !r[1]) {
 			*n = now - (d*60*60);
-			return 1;
+			goto ret;
 		}
 		if (errno == 0 && r[0] == 'm' && !r[1]) {
 			*n = now - (d*60);
-			return 1;
+			goto ret;
 		}
 		if (errno == 0 && r[0] == 's' && !r[1]) {
 			*n = now - d;
-			return 1;
+			goto ret;
 		}
-		parse_error("invalid relative time format '%s'", s-1);
+		parse_error("invalid relative time format '%s'", s);
 	}
 
 	parse_error("invalid time format '%s'", s);
 	return 0;
+ret:
+	free(s);
+	return 1;
 }
 
 static struct expr *
@@ -629,7 +895,7 @@ parse_timecmp()
 
 	op = parse_op();
 	if (!op)
-		parse_error("invalid comparison at '%.15s'", pos);
+		parse_error_at(NULL, "invalid comparison");
 
 	int64_t n;
 	if (parse_num(&n) || parse_dur(&n)) {
@@ -642,14 +908,43 @@ parse_timecmp()
 	return 0;
 }
 
+static struct expr *
+parse_redir(struct expr *e)
+{
+	char *s;
+	const char *m;
+
+	if (peek("||"))
+		return e;
+
+	if (token("|")) {
+		if (!parse_string(&s))
+			parse_error_at(NULL, "expected command");
+		struct expr *r = mkexpr(EXPR_REDIR_PIPE);
+		r->a.string = s;
+		r->b.string = xstrdup("w");
+		return chain(e, EXPR_AND, r);
+	}
+	else if (token(">>")) m = "a+";
+	else if (token(">")) m = "w+";
+	else return e;
+
+	if (!parse_string(&s))
+		parse_error_at(NULL, "expected file name");
+	struct expr *r = mkexpr(EXPR_REDIR_FILE);
+	r->a.string = s;
+	r->b.string = xstrdup(m);
+	return chain(e, EXPR_AND, r);
+}
+
 static struct expr *
 parse_and()
 {
-	struct expr *e1 = parse_timecmp();
+	struct expr *e1 = parse_redir(parse_timecmp());
 	struct expr *r = e1;
 
 	while (token("&&")) {
-		struct expr *e2 = parse_timecmp();
+		struct expr *e2 = parse_redir(parse_timecmp());
 		r = chain(r, EXPR_AND, e2);
 	}
 
@@ -671,116 +966,46 @@ parse_or()
 }
 
 static struct expr *
-parse_expr(char *s)
+parse_cond()
 {
-	pos = s;
-	struct expr *e = parse_or();
-	if (*pos)
-		parse_error("trailing garbage at '%.15s'", pos);
-	return e;
+	struct expr *e1 = parse_or();
+
+	if (token("?")) {
+		struct expr *e2 = parse_or();
+		if (token(":")) {
+			struct expr *e3 = parse_cond();
+			struct expr *r = mkexpr(EXPR_COND);
+			r->a.expr = e1;
+			r->b.expr = e2;
+			r->c.expr = e3;
+
+			return r;
+		} else {
+			parse_error_at(NULL, "expected :", pos);
+		}
+	}
+
+	return e1;
 }
 
 static struct expr *
-parse_msglist(char *s)
+parse_expr()
 {
-	int64_t n, m;
-	int r;
-	struct expr *e1, *e2;
-	char *d;
+	ws();
+	return parse_cond();
+}
 
-	switch (*s) {
-	case '/':
-		s++;
-		e1 = mkexpr(EXPR_REGEXI);
-		e1->a.string = "subject";
-		e1->b.regex = malloc(sizeof (regex_t));
-		r = regcomp(e1->b.regex, s, REG_EXTENDED | REG_NOSUB | REG_ICASE);
-		if (r != 0) {
-			char msg[256];
-			regerror(r, e1->b.regex, msg, sizeof msg);
-			parse_error("invalid regex '%s': %s", s, msg);
-		}
-		return e1;
-	case ':':
-		if (strlen(s) <= 1)
-			parse_error("missing type at '%.15s'", s);
-
-		enum flags flag;
-		n = 0;
-
-		switch (*++s) {
-		case 'P': flag = FLAG_PASSED; break;
-		case 'F': flag = FLAG_FLAGGED; break;
-		case 'D': flag = FLAG_DRAFT; break;
-		case 'd': /* FALL THROUGH */
-		case 'T': flag = FLAG_TRASHED; break;
-		case 'u': n = 1; /* FALL THROUGH */
-		case 'r': /* FALL THROUGH */
-		case 'S': flag = FLAG_SEEN; break;
-		case 'o': n = 1; /* FALL THROUGH */
-		case 'n': flag = FLAG_NEW; break;
-		case 'R': flag = FLAG_REPLIED; break;
-		default: parse_error("unknown type at '%.15s'", s);
-		}
-
-		e1 = mkexpr(EXPR_ANYSET);
-		e1->a.prop = PROP_FLAG;
-		e1->b.num = flag;
-
-		if (!n)
-			return e1;
-
-		e2 = mkexpr(EXPR_NOT);
-		e2->a.expr = e1;
-		return e2;
-	default:
-		pos = s;
-
-		if (((d = strchr(s, ':')) || (d = strchr(s, '-')))
-		    && parse_num(&n) && (pos = d + 1) && parse_num(&m)) {
-			/* index >= n */
-			e1 = mkexpr(EXPR_GE);
-			e1->a.prop = PROP_INDEX;
-			e1->b.num = n;
-
-			/* index <= m */
-			e2 = mkexpr(EXPR_LE);
-			e2->a.prop = PROP_INDEX;
-			e2->b.num = m;
-
-			/* e1 && e2 */
-			return chain(e1, EXPR_AND, e2);
-		} else if (parse_num(&n)) {
-			e1 = mkexpr(EXPR_EQ);
-			e1->a.prop = PROP_INDEX;
-			e1->b.num = n;
-
-			return e1;
-		} else {
-			char *disp, *addr;
-
-			blaze822_addr(s, &disp, &addr);
-			if (!disp && !addr)
-				parse_error("invalid address at '%.15s'", pos);
-
-			d = strdup((disp) ? disp : addr);
-
-			e1 = mkexpr(EXPR_REGEXI);
-			e1->a.prop = PROP_FROM;
-			e1->b.regex = malloc(sizeof (regex_t));
-			e1->extra = (disp) ? 0 : 1;
-
-			r = regcomp(e1->b.regex, d, REG_EXTENDED | REG_NOSUB | REG_ICASE);
-			if (r != 0) {
-				char msg[256];
-				regerror(r, e1->b.regex, msg, sizeof msg);
-				parse_error("invalid regex '%s': %s", d, msg);
-			}
-
-			return e1;
-		}
-	}
-	return 0;
+static struct expr *
+parse_buf(const char *f, char *s)
+{
+	struct expr *e;
+	fname = f;
+	line = pos = s;
+	linenr = 1;
+	e = parse_expr();
+	if (*pos)
+		parse_error_at(NULL, "trailing garbage");
+	return e;
 }
 
 time_t
@@ -789,8 +1014,13 @@ msg_date(struct mailinfo *m)
 	if (m->date)
 		return m->date;
 
-	if (!m->msg)
-		m->msg = blaze822(m->fpath);
+	// XXX: date comparisation should handle zero dates
+	if (!m->msg && m->fpath) {
+		if (!(m->msg = blaze822(m->fpath))) {
+			m->fpath = NULL;
+			return -1;
+		}
+	}
 
 	char *b;
 	if (m->msg && (b = blaze822_hdr(m->msg, "date")))
@@ -800,65 +1030,129 @@ msg_date(struct mailinfo *m)
 }
 
 char *
-msg_hdr(struct mailinfo *m, const char *h)
+msg_hdr(char **s, const char *h, struct mailinfo *m)
 {
-	if (!m->msg)
-		m->msg = blaze822(m->fpath);
+	static char hdrbuf[4096];
+
+	if (!m->msg && m->fpath) {
+		if (!(m->msg = blaze822(m->fpath))) {
+			m->fpath = NULL;
+			return NULL;
+		}
+	}
+
+	// XXX: only return one header for now
+	if (*s)
+		return NULL;
 
 	char *b;
 	if (!m->msg || !(b = blaze822_chdr(m->msg, h)))
-		goto err;
+		return NULL;
+	*s = b;
 
-	char buf[100];
-	blaze822_decode_rfc2047(buf, b, sizeof buf - 1, "UTF-8");
-	if (!*buf)
-		goto err;
-
-	return strdup(buf);
-err:
-	return "";
+	blaze822_decode_rfc2047(hdrbuf, b, sizeof hdrbuf - 1, "UTF-8");
+	return hdrbuf;
 }
 
 char *
-msg_addr(struct mailinfo *m, char *h, int t)
+msg_hdr_addr(char **s, const char *h, struct mailinfo *m, int rdisp)
 {
-	if (!m->msg)
-		m->msg = blaze822(m->fpath);
+	if (!m->msg && m->fpath) {
+		if (!(m->msg = blaze822(m->fpath))) {
+			m->fpath = NULL;
+			return NULL;
+		}
+	}
 
-	char *b;
-	if (m->msg == 0 || (b = blaze822_chdr(m->msg, h)) == 0)
-		return "";
+	char *b = *s;
+	if (!b) {
+		if (!m->msg || !(b = blaze822_chdr(m->msg, h)))
+			return NULL;
+	}
 
 	char *disp, *addr;
-	blaze822_addr(b, &disp, &addr);
+	*s = blaze822_addr(b, &disp, &addr);
 
-	if (t) {
-		if (!addr)
-			return "";
-		return addr;
-	} else {
-		if (!disp)
-			return "";
+	if (rdisp)
 		return disp;
+	return addr;
+}
+
+char *
+msg_hdr_address(char **s, const char *h, struct mailinfo *m)
+{
+	return msg_hdr_addr(s, h, m, 0);
+}
+
+char *
+msg_hdr_display(char **s, const char *h, struct mailinfo *m)
+{
+	return msg_hdr_addr(s, h, m, 1);
+}
+
+FILE *
+redir(struct expr *e)
+{
+	struct file *file;
+	FILE *fp;
+
+	for (file = files; file; file = file->next) {
+		if (e->op == file->op &&
+		    strcmp(e->a.string, file->name) == 0 &&
+			strcmp(e->b.string, file->mode) == 0)
+			return file->fp;
 	}
+
+	fflush(stdout);
+	fp = NULL;
+	switch (e->op) {
+	case EXPR_REDIR_FILE: fp = fopen(e->a.string, e->b.string); break;
+	case EXPR_REDIR_PIPE: fp = popen(e->a.string, e->b.string); break;
+	}
+	if (!fp) {
+		fprintf(stderr, "%s: %s: %s\n", argv0, e->a.string, strerror(errno));
+		exit(3);
+	}
+	file = xcalloc(1, sizeof (struct file));
+	file->op = e->op;
+	file->name = e->a.string;
+	file->mode = e->b.string;
+	file->fp = fp;
+	if (!files) files = file;
+	if (fileq) fileq->next = file;
+	fileq = file;
+	return fp;
 }
 
 int
 eval(struct expr *e, struct mailinfo *m)
 {
+	FILE *fp;
 	switch (e->op) {
 	case EXPR_OR:
 		return eval(e->a.expr, m) || eval(e->b.expr, m);
 	case EXPR_AND:
 		return eval(e->a.expr, m) && eval(e->b.expr, m);
+	case EXPR_COND:
+		return eval(e->a.expr, m)
+		    ? eval(e->b.expr, m)
+			: eval(e->c.expr, m);
 	case EXPR_NOT:
 		return !eval(e->a.expr, m);
-		return 1;
+	case EXPR_BINDING:
+		return eval(e->a.binding->expr, m);
 	case EXPR_PRUNE:
 		prune = 1;
 		return 1;
 	case EXPR_PRINT:
 		return 1;
+	case EXPR_REDIR_FILE:
+	case EXPR_REDIR_PIPE:
+		fp = redir(e);
+		fputs(m->file, fp);
+		putc('\n', fp);
+		fflush(fp);
+		return 1;
 	case EXPR_LT:
 	case EXPR_LE:
 	case EXPR_EQ:
@@ -869,15 +1163,15 @@ eval(struct expr *e, struct mailinfo *m)
 	case EXPR_ANYSET: {
 		long v = 0, n;
 
-		if (!m->sb && (
+		if (!m->sb && m->fpath && (
 		    e->a.prop == PROP_ATIME ||
 		    e->a.prop == PROP_CTIME ||
 		    e->a.prop == PROP_MTIME ||
-		    e->a.prop == PROP_SIZE) &&
-		    (m->sb = calloc(1, sizeof *m->sb)) &&
-		    stat(m->fpath, m->sb) != 0) {
-			fprintf(stderr, "stat");
-			exit(2);
+		    e->a.prop == PROP_SIZE)) {
+			m->sb = xcalloc(1, sizeof *m->sb);
+			if (stat(m->fpath, m->sb) == -1)
+				m->fpath = NULL;
+			// XXX: stat based expressions should handle 0
 		}
 
 		n = e->b.num;
@@ -924,77 +1218,79 @@ eval(struct expr *e, struct mailinfo *m)
 	case EXPR_GLOBI:
 	case EXPR_REGEX:
 	case EXPR_REGEXI: {
-		const char *s = "";
+		const char *s = NULL;
+		char *p = NULL;
+		char *(*fn)(char **, const char *, struct mailinfo *) = 0;
+		int rv = 0;
 		switch (e->a.prop) {
-		case PROP_PATH: s = m->fpath; break;
-		case PROP_FROM: s = msg_addr(m, "from", e->extra); break;
-		case PROP_TO: s = msg_addr(m, "to", e->extra); break;
-		default: s = msg_hdr(m, e->a.string); break;
+		case PROP_HDR: fn = msg_hdr; break;
+		case PROP_HDR_ADDR: fn = msg_hdr_address; break;
+		case PROP_HDR_DISP: fn = msg_hdr_display; break;
+		case PROP_PATH: s = m->fpath ? m->fpath : ""; break;
+		default: return 0;
 		}
-		switch (e->op) {
-		case EXPR_STREQ: return strcmp(e->b.string, s) == 0;
-		case EXPR_STREQI: return strcasecmp(e->b.string, s) == 0;
-		case EXPR_GLOB: return fnmatch(e->b.string, s, 0) == 0;
-		case EXPR_GLOBI: return fnmatch(e->b.string, s, FNM_CASEFOLD) == 0;
-		case EXPR_REGEX:
-		case EXPR_REGEXI: return regexec(e->b.regex, s, 0, 0, 0) == 0;
+		for (;;) {
+			if (fn && !(s = fn(&p, e->b.string, m)))
+				break;
+			switch (e->op) {
+			case EXPR_STREQ: rv = strcmp(e->c.string, s) == 0; break;
+			case EXPR_STREQI: rv = strcasecmp(e->c.string, s) == 0; break;
+			case EXPR_GLOB: rv = fnmatch(e->c.string, s, 0) == 0; break;
+			case EXPR_GLOBI:
+				rv = fnmatch(e->c.string, s, FNM_CASEFOLD) == 0; break;
+			case EXPR_REGEX:
+			case EXPR_REGEXI:
+				rv = regexec(e->c.regex, s, 0, 0, 0) == 0;
+				break;
+			}
+			if (!fn || rv) return rv;
 		}
+		return 0;
 	}
 	}
 	return 0;
 }
 
 struct mailinfo *
-mailfile(char *file)
+mailfile(struct mailinfo *m, char *file)
 {
 	static int init;
 	if (!init) {
-		// delay loading of the seqmap until we need to scan the first
-		// file, in case someone in the pipe updated the map before
-		char *seqmap = blaze822_seq_open(0);
-		blaze822_seq_load(seqmap);
+		// delay loading of the cur mail until we need to scan the first
+		// file, in case someone in the pipe updated it before
 		cur = blaze822_seq_cur();
 		init = 1;
 	}
+	char *fpath = file;
 
-	struct mailinfo *m;
-	m = calloc(1, sizeof *m);
-	if (!m) {
-		fprintf(stderr, "calloc");
-		exit(2);
-	}
-	m->fpath = file;
 	m->index = num++;
-	m->flags = 0;
-	m->replies = 0;
-	m->depth = 0;
-	m->sb = 0;
-	m->msg = 0;
+	m->file = file;
 
-	while (*m->fpath == ' ' || *m->fpath == '\t') {
+	while (*fpath == ' ' || *fpath == '\t') {
 		m->depth++;
-		m->fpath++;
+		fpath++;
 	}
 
-	char *e = m->fpath + strlen(m->fpath) - 1;
-	while (m->fpath < e && (*e == ' ' || *e == '\t'))
+	char *e = fpath + strlen(fpath) - 1;
+	while (fpath < e && (*e == ' ' || *e == '\t'))
 		*e-- = 0;
 
-	if (m->fpath[0] == '<') {
+	if (fpath[0] == '<') {
 		m->flags |= FLAG_SEEN | FLAG_INFO;
+		m->fpath = NULL;
 		return m;
 	}
 
-	if ((e = strrchr(m->fpath, '/') - 1) && (e - m->fpath) >= 2 &&
+	if ((e = strrchr(fpath, '/') - 1) && (e - fpath) >= 2 &&
 	    *e-- == 'w' && *e-- == 'e' && *e-- == 'n')
 		m->flags |= FLAG_NEW;
 
-	if (cur && strcmp(cur, m->fpath) == 0) {
+	if (cur && strcmp(cur, fpath) == 0) {
 		m->flags |= FLAG_CUR;
 		cur_idx = m->index;
 	}
 
-	char *f = strstr(m->fpath, ":2,");
+	char *f = strstr(fpath, ":2,");
 	if (f) {
 		if (strchr(f, 'P'))
 			m->flags |= FLAG_PASSED;
@@ -1010,6 +1306,7 @@ mailfile(char *file)
 			m->flags |= FLAG_FLAGGED;
 	}
 
+	m->fpath = fpath;
 	return m;
 }
 
@@ -1038,12 +1335,8 @@ do_thr()
 			break;
 
 		if (((Tflag && thr->matched) || ml->m->matched) && !ml->m->prune) {
-			int i;
-			for (i = 0; i < ml->m->depth; i++)
-				putchar(' ');
-
-			fputs(ml->m->fpath, stdout);
-			putchar('\n');
+			fputs(ml->m->file, stdout);
+			putc('\n', stdout);
 
 			kept++;
 		}
@@ -1055,7 +1348,7 @@ do_thr()
 		if (ml->m->sb)
 			free(ml->m->sb);
 
-		free(ml->m->fpath);
+		free(ml->m->file);
 		free(ml->m);
 	}
 
@@ -1068,21 +1361,19 @@ collect(char *file)
 {
 	struct mailinfo *m;
 	struct mlist *ml;
+	char *f;
 
-	if ((m = mailfile(file)) == 0)
-		return;
+	f = xstrdup(file);
+	m = xcalloc(1, sizeof *m);
+	m = mailfile(m, f);
 
 	if (m->depth == 0) {
+		/* process previous thread */
 		if (thr)
 			do_thr();
 
 		/* new thread */
-		thr = calloc(1, sizeof *thr);
-		if (!thr) {
-			fprintf(stderr, "calloc");
-			exit(2);
-		}
-
+		thr = xcalloc(1, sizeof *thr);
 		thr->matched = 0;
 		ml = thr->cur = thr->childs;
 		thr->cur->m = m;
@@ -1112,29 +1403,27 @@ collect(char *file)
 
 	for (ml = ml->parent; ml; ml = ml->parent)
 		ml->m->replies++;
-
-	m->fpath = strdup(m->fpath);
 }
 
 void
 oneline(char *file)
 {
-	struct mailinfo *m;
-
-	m = mailfile(file);
-	if (expr && !eval(expr, m))
+	struct mailinfo m = { 0 };
+	m.index = num++;
+	(void) mailfile(&m, file);
+	if (expr && !eval(expr, &m))
 		goto out;
 
 	fputs(file, stdout);
-	putchar('\n');
+	putc('\n', stdout);
+	fflush(stdout);
 	kept++;
 
 out:
-	if (m->msg)
-		blaze822_free(m->msg);
-	if (m->sb)
-		free(m->sb);
-	free(m);
+	if (m.msg)
+		blaze822_free(m.msg);
+	if (m.sb)
+		free(m.sb);
 }
 
 int
@@ -1142,33 +1431,62 @@ main(int argc, char *argv[])
 {
 	long i;
 	int c;
+	int vflag;
 
 	argv0 = argv[0];
 	now = time(0);
 	num = 1;
+	vflag = 0;
 
-	while ((c = getopt(argc, argv, "Tt:")) != -1)
+	while ((c = getopt(argc, argv, "F:Tt:v")) != -1)
 		switch (c) {
+		case 'F':
+		{
+			char *s;
+			off_t len;
+			int r = slurp(optarg, &s, &len);
+			if (r != 0) {
+				fprintf(stderr, "%s: error opening file '%s': %s\n",
+				    argv0, optarg, strerror(r));
+				exit(1);
+			}
+			expr = chain(expr, EXPR_AND, parse_buf(optarg, s));
+			free(s);
+			break;
+		}
 		case 'T': Tflag = need_thr = 1; break;
-		case 't': expr = chain(expr, EXPR_AND, parse_expr(optarg)); break;
+		case 't': expr = chain(expr, EXPR_AND, parse_buf("argv", optarg)); break;
+		case 'v': vflag = 1; break;
 		default:
-			fprintf(stderr, "Usage: %s [-T] [-t test] [msglist ...]\n", argv0);
+			fprintf(stderr, "Usage: %s [-Tv] [-t test] [-F file] [msgs...]\n", argv0);
 			exit(1);
 		}
 
-	if (optind != argc)
-		for (c = optind; c < argc; c++)
-			expr = chain(expr, EXPR_AND, parse_msglist(argv[c]));
+	xpledge("stdio rpath wpath cpath proc exec", 0);
 
-	if (isatty(0))
-		i = blaze822_loop1(":", need_thr ? collect : oneline);
+	void (*cb)(char *) = need_thr ? collect : oneline;
+	if (argc == optind && isatty(0))
+		i = blaze822_loop1(":", cb);
 	else
-		i = blaze822_loop(0, 0, need_thr ? collect : oneline);
+		i = blaze822_loop(argc-optind, argv+optind, cb);
 
 	/* print and free last thread */
 	if (Tflag && thr)
 		do_thr();
 
-	fprintf(stderr, "%ld mails tested, %ld picked.\n", i, kept);
+	freeexpr(expr);
+
+	if (vflag)
+		fprintf(stderr, "%ld mails tested, %ld picked.\n", i, kept);
+
+	for (; files; files = fileq) {
+		fileq = files->next;
+		if (files->op == EXPR_REDIR_PIPE)
+			pclose(files->fp);
+		else
+			fclose(files->fp);
+		free(files);
+	}
+
 	return 0;
 }
