@@ -11,6 +11,7 @@
 #include <unistd.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 static char *expr;
 
@@ -28,7 +29,7 @@ subst(char *str, char *srch, char *repl, char *flags)
 #define APPC(c) do { if (b >= bufe) return str; *b++ = c; } while (0)
 
 	regex_t srchrx;
-	regmatch_t pmatch[10];
+	regmatch_t pmatch[10] = {{0}};
 	if (regcomp(&srchrx, srch, iflag ? REG_ICASE : 0) != 0)
 		return str;
 
@@ -137,7 +138,7 @@ sed(char *file)
 					if (*s == ':')
 						*s = '|';
 				int rv;
-				if ((rv = regcomp(&headerrx, headersel, REG_EXTENDED)) != 0) {
+				if ((rv = regcomp(&headerrx, headersel, REG_EXTENDED | REG_ICASE)) != 0) {
 					char buf[100];
 					regerror(rv, &headerrx, buf, sizeof buf);
 					fprintf(stderr, "msed: %s\n", buf);
@@ -217,6 +218,8 @@ sed(char *file)
 					free(to);
 					free(flags);
 
+					break;
+				case '\0':
 					break;
 				default:
 					fprintf(stderr, "msed: unknown command: '%c'\n", *e);
@@ -321,6 +324,8 @@ main(int argc, char *argv[])
 			exit(1);
 		}
 
+	xpledge("stdio rpath", "");
+
 	expr = argv[optind];
 	optind++;
 
