@@ -14,6 +14,7 @@
 #include <unistd.h>
 
 #include "blaze822.h"
+#include "xpledge.h"
 
 static int Bflag;
 static int rflag;
@@ -35,6 +36,7 @@ static char fallback_ct[] = "text/plain";
 struct message *filters;
 
 static int mimecount;
+static int extractcount;
 static int safe_output;
 
 static int reply_found;
@@ -49,10 +51,10 @@ printable(int c)
 }
 
 size_t
-print_ascii(char *body, size_t bodylen)
+print_ascii(char *body, size_t bodylen, int oneline)
 {
 	if (safe_output) {
-		safe_u8putstr(body, bodylen, stdout);
+		safe_u8putstr(body, bodylen, oneline, stdout);
 		return bodylen;
 	} else {
 		return fwrite(body, 1, bodylen, stdout);
@@ -71,7 +73,7 @@ printhdr(char *hdr)
 	}
 
 	if (*hdr) {
-		print_ascii(hdr, strlen(hdr));
+		print_ascii(hdr, strlen(hdr), 1);
 		fputc('\n', stdout);
 	}
 }
@@ -83,7 +85,7 @@ print_u8recode(char *body, size_t bodylen, char *srcenc)
 
 	ic = iconv_open("UTF-8", srcenc);
 	if (ic == (iconv_t)-1) {
-		printf("unsupported encoding: %s\n", srcenc);
+		fprintf(stderr, "unsupported encoding: %s\n", srcenc);
 		return;
 	}
 
@@ -96,7 +98,7 @@ print_u8recode(char *body, size_t bodylen, char *srcenc)
 		size_t r = iconv(ic, &body, &bodylen, &bufptr, &buflen);
 
 		if (bufptr != buf) {
-			print_ascii(buf, bufptr-buf);
+			print_ascii(buf, bufptr-buf, 0);
 			final_char = bufptr[-1];
 		}
 
@@ -105,7 +107,7 @@ print_u8recode(char *body, size_t bodylen, char *srcenc)
 			buflen = sizeof buf;
 			r = iconv(ic, 0, 0, &bufptr, &buflen);
 			if (bufptr != buf) {
-				print_ascii(buf, bufptr-buf);
+				print_ascii(buf, bufptr-buf, 0);
 				final_char = bufptr[-1];
 			}
 			if (r != (size_t)-1)
@@ -183,11 +185,23 @@ print_filename(char *filename)
 {
 	if (filename) {
 		printf(" name=\"");
-		safe_u8putstr(filename, strlen(filename), stdout);
+		safe_u8putstr(filename, strlen(filename), 1, stdout);
 		printf("\"");
 	}
 }
 
+static int
+probably_utf8(char *charset)
+{
+	return !charset ||
+	    strcasecmp(charset, "utf-8") == 0 ||
+	    strcasecmp(charset, "utf8") == 0 ||
+	    strcasecmp(charset, "us-ascii") == 0 ||
+	    /* best guess for: */
+	    strcasecmp(charset, "unknown-8bit") == 0 ||
+	    strcasecmp(charset, "x-unknown") == 0;
+}
+
 blaze822_mime_action
 render_mime(int depth, struct message *msg, char *body, size_t bodylen)
 {
@@ -224,6 +238,11 @@ render_mime(int depth, struct message *msg, char *body, size_t bodylen)
 		}
 		setenv("PIPE_CONTENTTYPE", ct, 1);
 
+		char *messageid = blaze822_hdr(msg, "message-id");
+		if (messageid) {
+			setenv("MESSAGE_ID", messageid, 1);
+		}
+
 		char *output;
 		size_t outlen;
 		int e = filter(body, bodylen, cmd, &output, &outlen);
@@ -233,7 +252,7 @@ render_mime(int depth, struct message *msg, char *body, size_t bodylen)
 				printf(" render=\"%s\" ---\n", cmd);
 			if (outlen) {
 				if (e == 0)
-					print_ascii(output, outlen);
+					print_ascii(output, outlen, 0);
 				else
 					fwrite(output, 1, outlen, stdout);
 				if (output[outlen-1] != '\n')
@@ -279,11 +298,8 @@ nofilter:
 			char *charset = 0, *cs, *cse;
 			if (blaze822_mime_parameter(ct, "charset", &cs, &cse))
 				charset = strndup(cs, cse-cs);
-			if (!charset ||
-			    strcasecmp(charset, "utf-8") == 0 ||
-			    strcasecmp(charset, "utf8") == 0 ||
-			    strcasecmp(charset, "us-ascii") == 0) {
-				print_ascii(body, bodylen);
+			if (probably_utf8(charset)) {
+				print_ascii(body, bodylen, 0);
 				if (bodylen > 0 && body[bodylen-1] != '\n')
 					putchar('\n');
 			} else {
@@ -300,9 +316,10 @@ nofilter:
 			}
 			printf("\n");
 		} else if (strncmp(ct, "multipart/alternative", 21) == 0) {
-			choose_alternative(msg, depth);
-
-			r = MIME_PRUNE;
+			if (strcmp(Aflag, "all") != 0) {
+				choose_alternative(msg, depth);
+				r = MIME_PRUNE;
+			} // else default blaze822_mime_walk action
 		} else if (strncmp(ct, "multipart/", 10) == 0) {
 			; // default blaze822_mime_walk action
 		} else {
@@ -319,13 +336,17 @@ nofilter:
 static void
 choose_alternative(struct message *msg, int depth)
 {
-	int n = 1;
+	int n = 0;
 	int m = 0;
 	char *p = Aflag + strlen(Aflag);
 
 	struct message *imsg = 0;
 	while (blaze822_multipart(msg, &imsg)) {
 		m++;
+
+		if (blaze822_bodylen(imsg) == 0)
+			continue;
+
 		char *ict = blaze822_hdr(imsg, "content-type");
 		if (!ict)
 			ict = fallback_ct;
@@ -342,6 +363,13 @@ choose_alternative(struct message *msg, int depth)
 	}
 	blaze822_free(imsg);
 
+	if (n == 0) {
+		// No part matched, use last part as per RFC1341 7.2.3
+		// "In general, choosing the best type means displaying
+		// only the LAST part that can be displayed."
+		n = m;
+	}
+
 	imsg = 0;
 	while (blaze822_multipart(msg, &imsg))
 		if (--n == 0)
@@ -362,11 +390,8 @@ reply_mime(int depth, struct message *msg, char *body, size_t bodylen)
 		char *charset = 0, *cs, *cse;
 		if (blaze822_mime_parameter(ct, "charset", &cs, &cse))
 			charset = strndup(cs, cse-cs);
-		if (!charset ||
-		    strcasecmp(charset, "utf-8") == 0 ||
-		    strcasecmp(charset, "utf8") == 0 ||
-		    strcasecmp(charset, "us-ascii") == 0)
-			print_ascii(body, bodylen);
+		if (probably_utf8(charset))
+			print_ascii(body, bodylen, 0);
 		else
 			print_u8recode(body, bodylen, charset);
 		reply_found++;
@@ -477,7 +502,7 @@ extract_mime(int depth, struct message *msg, char *body, size_t bodylen)
 			fwrite(body, 1, bodylen, stdout);
 		} else { // extract all named attachments
 			if (filename) {
-				safe_u8putstr(filename, strlen(filename), stdout);
+				safe_u8putstr(filename, strlen(filename), 1, stdout);
 				printf("\n");
 				writefile(filename, body, bodylen);
 			}
@@ -521,6 +546,7 @@ extract_mime(int depth, struct message *msg, char *body, size_t bodylen)
 					printf("%s\n", bufptr);
 					writefile(bufptr, body, bodylen);
 				}
+				extractcount++;
 			} else if (filename &&
 			    fnmatch(a, filename, FNM_PATHNAME) == 0) {
 				// extract by name
@@ -537,10 +563,11 @@ extract_mime(int depth, struct message *msg, char *body, size_t bodylen)
 						fwrite(body, 1, bodylen, stdout);
 					}
 				} else {
-					safe_u8putstr(filename, strlen(filename), stdout);
+					safe_u8putstr(filename, strlen(filename), 1, stdout);
 					printf("\n");
 					writefile(filename, body, bodylen);
 				}
+				extractcount++;
 			}
 		}
 	}
@@ -551,6 +578,8 @@ extract_mime(int depth, struct message *msg, char *body, size_t bodylen)
 void
 extract_cb(char *file)
 {
+	while (*file == ' ' || *file == '\t')
+		file++;
 	struct message *msg = blaze822_file(file);
 	if (!msg)
 		return;
@@ -580,7 +609,7 @@ print_date_header(char *v)
 	}
 
 	printf("Date: ");
-	print_ascii(v, strlen(v));
+	print_ascii(v, strlen(v), 1);
 
 	time_t t = blaze822_date(v);
 	if (t == -1) {
@@ -666,7 +695,7 @@ print_decode_header(char *h, char *v)
 	printhdr(h);
 	fputc(':', stdout);
 	fputc(' ', stdout);
-	print_ascii(d, strlen(d));
+	print_ascii(d, strlen(d), 1);
 	fputc('\n', stdout);
 }
 
@@ -733,7 +762,7 @@ show(char *file)
 	printf("\n");
 
 	if (rflag) {  // raw body
-		print_ascii(blaze822_body(msg), blaze822_bodylen(msg));
+		print_ascii(blaze822_body(msg), blaze822_bodylen(msg), 0);
 		goto done;
 	}
 
@@ -777,7 +806,9 @@ main(int argc, char *argv[])
 			exit(1);
 		}
 
-	if (!rflag && !Oflag && !Rflag)
+	xpledge("stdio rpath wpath cpath proc exec", NULL);
+
+	if (!rflag && !xflag && !Oflag && !Rflag)
 		safe_output = 1;
 
 	if (safe_output && isatty(1)) {
@@ -786,7 +817,7 @@ main(int argc, char *argv[])
 		if (!pg) {
 			pg = getenv("PAGER");
 			if (pg && strcmp(pg, "less") == 0) {
-				static char lesscmd[] = "less -RFXe";
+				static char lesscmd[] = "less -RFX";
 				pg = lesscmd;
 			}
 		}
@@ -802,17 +833,22 @@ main(int argc, char *argv[])
 	}
 
 	if (xflag) { // extract
+		xpledge("stdio rpath wpath cpath", NULL);
 		extract(xflag, argc-optind, argv+optind, 0);
 	} else if (Oflag) { // extract to stdout
+		xpledge("stdio rpath", NULL);
 		extract(Oflag, argc-optind, argv+optind, 1);
 	} else if (tflag) { // list
+		xpledge("stdio rpath", NULL);
 		if (argc == optind && isatty(0))
 			blaze822_loop1(".", list);
 		else
 			blaze822_loop(argc-optind, argv+optind, list);
 	} else if (Rflag) { // render for reply
+		xpledge("stdio rpath", NULL);
 		blaze822_loop(argc-optind, argv+optind, reply);
 	} else { // show
+		/* XXX pledge: still r/w on the whole file-system + fork/exec */
 		if (!(qflag || rflag || Fflag)) {
 			char *f = getenv("MAILFILTER");
 			if (!f)
@@ -825,7 +861,8 @@ main(int argc, char *argv[])
 		else
 			blaze822_loop(argc-optind, argv+optind, show);
 		if (!nflag) // don't set cur
-			blaze822_seq_setcur(newcur);
+			if (newcur)
+				blaze822_seq_setcur(newcur);
 	}
 
 	if (pid2 > 0)
@@ -833,5 +870,10 @@ main(int argc, char *argv[])
 	if (pid1 > 0)
 		pipeclose(pid1);
 
+	if ((xflag || Oflag) && extractcount < argc-optind) {
+		fprintf(stderr, "mshow: could not extract all attachments\n");
+		return 1;
+	}
+
 	return 0;
 }
